﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Json;
using QuickGraph;
using QuickGraph.Algorithms;
using QuickGraph.Algorithms.Observers;
using QuickGraph.Algorithms.ShortestPath;

//
// OSM Router sample
//
// Written by Bjarke Viksoe (bjarke@viksoe.dk)
// Copyright (c) 2015 Bjarke Viksoe.
//
// This file is provided "as is" with no expressed or implied warranty.
// The author accepts no liability if it causes any damage to you or your
// computer whatsoever. It's free, so don't hassle me about it.
//
// Beware of bugs.
// 
// Data extract from here:  http://overpass-turbo.eu/
//   [out:json][timeout:60];
//   (
//     relation["route"="subway"]({{bbox}});
//   );
//   >>;
//   out body;
// 


namespace OsmwayRouter
{
    [DataContract]
    class OsmOverpass
    {
        [DataMember(IsRequired=true)] 
        public string version = null;
        [DataMember(IsRequired = true)] 
        public string generator = null;
        [DataMember(IsRequired = true)]
        public List<OsmElement> elements = null;
    }

    [DataContract]
    public class OsmElement
    {
        [DataMember(IsRequired = true)] 
        public string type = null;
        [DataMember(IsRequired = true)]
        public ulong id;
        [DataMember]
        public double lat;
        [DataMember]
        public double lon;
        [DataMember]
        public OsmTags tags = null;
        [DataMember]
        public List<ulong> nodes = null;
        [DataMember]
        public List<OsmMember> members = null;

        public List<Vertex> shared = null;
    }

    [DataContract]
    public class OsmMember
    {
        [DataMember(IsRequired = true)]
        public string type = null;
        [DataMember(Name = "ref", IsRequired = true)]
        public ulong reference;
        [DataMember]
        public string role = "";
    }

    [DataContract]
    public class OsmTags
    {
        [DataMember(Name = "name", EmitDefaultValue = true)]
        public string name = null;
        [DataMember(Name="name:en")]
        public string nameEnglish = null;
        [DataMember]
        public string highway = null;
        [DataMember]
        public string station = null;
        [DataMember]
        public string colour = null;
    }

    public class Vertex
    {
        public OsmElement el { get; set; }
        public OsmElement rel { get; set; }

        public override string ToString()
        {
            if (el != null && el.tags != null && el.tags.nameEnglish != null) return el.tags.nameEnglish + " (" +  el.type + ") " + rel.tags.nameEnglish;
            if (el != null && el.tags != null && el.tags.name != null) return el.tags.name + " (" + el.type + ") " + rel.tags.name;
            return base.ToString();
        }
    };

    class Program
    {
        static void Main(string[] args)
        {
            var routes = new Dictionary<ulong, OsmElement>();
            var nodes = new Dictionary<ulong, OsmElement>();
            var ways = new Dictionary<ulong, OsmElement>();

            var stream = File.Open(@"Assets\export.json", FileMode.Open, FileAccess.Read, FileShare.Read);
            var js = new DataContractJsonSerializer(typeof(OsmOverpass));
            OsmOverpass import = js.ReadObject(stream) as OsmOverpass;
            stream.Close();

            var hash = new HashSet<ulong>();
            foreach (var el in import.elements)
            {
                if (hash.Contains(el.id))
                {
                    System.Diagnostics.Debug.Assert(false, "Import have duplicated elements");
                    continue;
                }
                if (el.type == "relation") routes[el.id] = el;
                if (el.type == "node") nodes[el.id] = el;
                if (el.type == "way") ways[el.id] = el;
                hash.Add(el.id);
            }

            double locBeginLat = 37.558198, locBeginLon = 126.936794;  // Sinchon
            
            double locDestLat = 37.505737, locDestLon = 126.939289;  // Jangseungbaegi
            //double locDestLat = 37.525333, locDestLon = 126.884980;  // Yangpuong
            //double locDestLat = 37.550705, locDestLon = 126.914289;  // Hapjeong
            //double locDestLat = 37.513773, locDestLon = 127.099972;  // Lotte
            //double locDestLat = 37.5183, locDestLon = 127.1246;      // Olympic Park

            var beginVertex = new Vertex()
            {
                rel = new OsmElement() { id = 1, type = "relation", tags = new OsmTags() { name = "Begin Journey", nameEnglish = "Begin Journey" } },
                el = new OsmElement() { id = 2, lat = locBeginLat, lon = locBeginLon, type = "point-begin", tags = new OsmTags() { name = "Begin Node", nameEnglish = "Begin Node" } },
            };
            var endVertex = new Vertex() 
            {
                rel = new OsmElement() { id = 3, type = "relation", tags = new OsmTags() { name = "End Journey", nameEnglish = "End Journey" } },
                el = new OsmElement() { id = 4, lat = locDestLat, lon = locDestLon, type = "point-end", tags = new OsmTags() { name = "End Node", nameEnglish = "End Node" } },
            };

            var graph = new Dictionary<Vertex, List<Vertex>>();

            foreach (var route in routes.Values)
            {
                OsmElement way;
                bool bothDirections = false;
                foreach (var member in route.members.Where(e => e.type == "way"))
                {
                    if (member.role == "") bothDirections = true;
                    if (member.role == "route") bothDirections = true;
                    if (!ways.TryGetValue(member.reference, out way)) continue;
                    route.tags.colour = route.tags.colour ?? way.tags.colour;
                }

                Vertex previousVertex = null;
                foreach (var member in route.members.Where(e => e.role == "stop"))
                {
                    OsmElement element = null;
                    if (!nodes.TryGetValue(member.reference, out element)) continue;

                    var vertex = new Vertex() { rel = route, el = element };

                    graph[vertex] = new List<Vertex>();
                    
                    if (vertex.el.shared == null) vertex.el.shared = new List<Vertex>();
                    vertex.el.shared.Add(vertex);

                    if (previousVertex != null)
                    {
                        if (bothDirections) graph[vertex].Add(previousVertex);
                        graph[previousVertex].Add(vertex);
                    }
                    previousVertex = vertex;
                }
            }

            // Add shared elements (other relations that have same stop in their route)...
            foreach (var vertex in graph.Keys)
            {
                foreach (var n in vertex.el.shared)
                    if (vertex != n && !graph[vertex].Contains(n)) graph[vertex].Add(n);
            }

            // Add near-by station within walking distance...
            foreach (var vertex in graph.Keys)
            {
                foreach (var n in graph.Keys)
                {
                    if (n.rel == vertex.rel) continue;
                    if (!String.IsNullOrEmpty(n.el.tags.station) && n.el.tags.station == vertex.el.tags.station) continue;
                    if (DistanceEuclidean(n.el, vertex.el) > 0.800) continue;
                    if (!graph[vertex].Contains(n)) graph[vertex].Add(n);
                }
            }

            graph[beginVertex] = new List<Vertex>();
            graph[endVertex] = new List<Vertex>();

            bool beginConnected = false, endConnected = false;
            foreach (var vertex in graph.Keys)
            {
                if (DistanceEuclidean(vertex.el, beginVertex.el) < 1.0)
                {
                    graph[beginVertex].Add(vertex);
                    beginConnected = true;
                }
                if (DistanceEuclidean(vertex.el, endVertex.el) < 1.0)
                {
                    graph[vertex].Add(endVertex);
                    endConnected = true;
                }
            }

            // Make sure at least one vertex is connected to the edges...
            if (!beginConnected)
                graph[beginVertex].Add(FindClosestRoutableVertex(graph, beginVertex));
            if (!endConnected)
                graph[FindClosestRoutableVertex(graph, endVertex)].Add(endVertex);

            Func<SEquatableEdge<Vertex>, double> edgeCost = (edge) =>
            {
                Cost cost = DetermineCost(edge.Source, edge.Target);
                return (cost.distance / cost.speed) + (cost.idle / 60.0);
            };
            Func<Vertex, double> heuristicCost = (n) =>
            {
                return DistanceEuclidean(n.el, endVertex.el);
            };

            var velg = graph.ToVertexAndEdgeListGraph(kv => kv.Value.Select(n => new SEquatableEdge<Vertex>(kv.Key, n)));
            
            var algo = new DijkstraShortestPathAlgorithm<Vertex, SEquatableEdge<Vertex>>(velg, edgeCost);
            //var algo = new BellmanFordShortestPathAlgorithm<Vertex, SEquatableEdge<Vertex>>(velg, edgeCost);
            //var algo = new AStarShortestPathAlgorithm<Vertex, SEquatableEdge<Vertex>>(velg, edgeCost, heuristicCost);
            var vis = new VertexPredecessorRecorderObserver<Vertex, SEquatableEdge<Vertex>>();
            using (vis.Attach(algo))
            {
                algo.Compute(beginVertex);
                IEnumerable<SEquatableEdge<Vertex>> path;
                if (vis.TryGetPath(endVertex, out path))
                    foreach (var n in path)
                        Console.WriteLine(n.ToString());
            }

            /*
            var algo = new FloydWarshallAllShortestPathAlgorithm<Vertex, SEquatableEdge<Vertex>>(velg, edgeCost);
            algo.Compute();
            IEnumerable<SEquatableEdge<Vertex>> path;
            if (algo.TryGetPath(beginVertex, endVertex, out path))
                foreach (var n in path)
                    Console.WriteLine(n.ToString());
            */
        }

        private class Cost
        {
            public double distance;
            public double speed;
            public double idle;
        }

        private static Cost DetermineCost(Vertex start, Vertex end)
        {
            Cost cost = new Cost();
            cost.distance = DistanceEuclidean(start.el, end.el);
            cost.speed = cost.distance > 1.2 ? 40.0 : 27.0;                // Average speed of subway train
            if (start.rel != end.rel)
            {
                cost.speed = 5.0;                                          // Preferred walking speed
                cost.idle = 8.0;                                           // Minutes waiting for next connection
            }
            return cost;
        }

        private static Vertex FindClosestRoutableVertex(IDictionary<Vertex, List<Vertex>> graph, Vertex vertex)
        {
            return graph.Keys.Aggregate((c1, c2) => DistanceEuclidean(vertex.el, c1.el) < DistanceEuclidean(vertex.el, c2.el) ? c1 : c2);
        }

        private static double Distance2D(OsmElement start, OsmElement end)
        {
            return Distance2D(start.lat, start.lon, end.lat, end.lon);
        }

        private static double DistanceHaversine(OsmElement start, OsmElement end)
        {
            return DistanceHaversine(start.lat, start.lon, end.lat, end.lon);
        }

        private static double DistanceEuclidean(OsmElement start, OsmElement end)
        {
            return DistanceEuclidean(start.lat, start.lon, end.lat, end.lon);
        }

        const double DEG2RAD = Math.PI / 180.0;
        const double DEGLEN = 110.25;
        const double EARTH_RADIUS_KM = 6371.0;

        /// <summary>
        /// Crude inaccurate distance calculation between two coordinates.
        /// </summary>
        private static double Distance2D(double lat1, double lon1, double lat2, double lon2)
        {
            double dx = (lat2 - lat1);
            double dy = (lon2 - lon1);
            return Math.Sqrt(dx * dx + dy * dy) * DEGLEN;
        }

        /// <summary>
        /// Accurate distance calculation between two coordinates.
        /// </summary>
        private static double DistanceHaversine(double lat1, double lon1, double lat2, double lon2)
        {
            // http://damien.dennehy.me/blog/2011/01/15/haversine-algorithm-in-csharp/
            double dLat = DEG2RAD * (lat2 - lat1);
            double dLon = DEG2RAD * (lon2 - lon1);
            double a = Math.Pow(Math.Sin(dLat / 2.0), 2.0) + Math.Cos(DEG2RAD * (lat1)) * Math.Cos(DEG2RAD * (lat2)) * Math.Pow(Math.Sin(dLon / 2.0), 2.0);
            double c = 2.0 * Math.Atan2(Math.Sqrt(a), Math.Sqrt(1.0 - a));
            return EARTH_RADIUS_KM * c;
        }

        /// <summary>
        /// Approximated distance calculation between two coordinates.
        /// Use only for smaller distances.
        /// </summary>
        private static double DistanceEuclidean(double lat1, double lon1, double lat2, double lon2)
        {
            // http://www.movable-type.co.uk/scripts/latlong.html
            lat1 = DEG2RAD * (lat1); lon1 = DEG2RAD * (lon1);
            lat2 = DEG2RAD * (lat2); lon2 = DEG2RAD * (lon2);
            var x = (lon2 - lon1) * Math.Cos((lat1 + lat2) / 2);
            var y = (lat2 - lat1);
            return Math.Sqrt(x * x + y * y) * EARTH_RADIUS_KM;
        }
    }
}
